# from flask_appbuilder.models.mixins import AuditMixin
from sqlalchemy import Column, Text, Integer, ForeignKey
# from sqlalchemy.ext.declarative import declarative_base
#
# Base = declarative_base()
from flask_appbuilder import Model
from sqlalchemy.orm import relationship


class docModel(Model):
    __tablename__ = "docs"
    id = Column(Integer, primary_key=True, autoincrement=False)
    name = Column(Text())
    doc = Column(Text(), default="")
    names = Column(Text(), default="")
    level = Column(Integer(), default=-1)
    source = Column(Text(), default="")
    path = Column(Text(), default="")
    code = Column(Text(), default="")
    ctype = Column(Text(), default="")
    alias = Column(Text(), default="")

    def __repr__(self):
        return f"<{self.name}>"

    def new_parent(self, parent):
        if parent:
            self.parents.append(parent)
            parent.kids.append(self)

    def new_name(self, nname):
        self.names.append(nname)

    def to_dicts(self, *cols):
        return dict((col, getattr(self, col)) for col in cols)


class docGraphModel(Model):
    __tablename__ = "doc_graph"
    id = Column(Integer, primary_key=True)
    parent_id = Column(Integer(), ForeignKey("docs.id"))
    parent = relationship(docModel, foreign_keys=[parent_id])
    kid_id = Column(Integer(), ForeignKey("docs.id"))
    kid = relationship(docModel, foreign_keys=[kid_id])

class inhGraphModel(Model):
    __tablename__ = "inh_graph"
    id = Column(Integer, primary_key=True)
    anc_id = Column(Integer(), ForeignKey("docs.id")) # ancestor
    anc = relationship(docModel, foreign_keys=[anc_id])
    des_id = Column(Integer(), ForeignKey("docs.id")) # descendant
    des = relationship(docModel, foreign_keys=[des_id])


# Mapping the self-relate docs on attribute relations
docModel.kids = relationship(docModel,
                             secondary="doc_graph",
                             primaryjoin=(docModel.id == docGraphModel.parent_id),
                             secondaryjoin=(docGraphModel.kid_id == docModel.id)
                             )
docModel.parents = relationship(docModel,
                                secondary="doc_graph",
                                primaryjoin=(docModel.id == docGraphModel.kid_id),
                                secondaryjoin=(docGraphModel.parent_id == docModel.id)
                                )

# Mappping the self-relate docs on class inheritance
docModel.dess = relationship(docModel,
                             secondary="inh_graph",
                             primaryjoin=(docModel.id == inhGraphModel.anc_id),
                             secondaryjoin=(inhGraphModel.des_id == docModel.id)
                             )
docModel.ancs = relationship(docModel,
                                secondary="inh_graph",
                                primaryjoin=(docModel.id == inhGraphModel.des_id),
                                secondaryjoin=(inhGraphModel.anc_id == docModel.id)
                                )


from flask_appbuilder import BaseView, expose, ModelView
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask import jsonify, request
import json
import inspect
import logging
import importlib
from doctour.base.parse import docTour
from doctour.base.doc import docModel,docGraphModel,inhGraphModel
from doctour.app.models.lib import libModel

from sqlalchemy.orm.session import sessionmaker
from sqlalchemy import create_engine as ce

Session = sessionmaker()


class searchView(BaseView):
    route_base = "/search"

    @expose("/index/", methods=["POST"])
    def major_view(self):
        data_ = json.loads(request.data)
        kw = data_["kw"]
        logging.info(kw)
        try:
            logging.info(f"loading {kw}")
            obj = __import__(kw)
            logging.info(f"{str(obj)} loaded")
        except Exception as e:
            return jsonify({
                "status": 500, "success": False, "data": {"error": str(e)}
            })
        obj_info = dict({
            "name": kw,
            "file": inspect.getsourcefile(obj),
        })

        return jsonify({
            "status": 200, "success": True, "data": obj_info
        })

class docView(ModelView):
    route_base = "/doc"
    datamodel = SQLAInterface(docModel)

    @expose("/read/<lib>/<doc_id>/", methods=["GET"])
    def read_doc(self, lib, doc_id):
        return self.render_template("read_doc.html")

    def item_dict(self,item):
        return dict({"id":item.id, "name":item.name,"alias": item.alias,"ctype":item.ctype})

    def get_lib(self,libname):
        lib_item = self.appbuilder.session.query(libModel).filter(libModel.name == libname).first()
        return lib_item

    @expose("/traceup/<lib>/<doc_id>/", methods=["POST","GET"])
    def trace_up(self, lib, doc_id):
        lib_item = self.get_lib(lib)
        eng = ce(lib_item.data)
        sess = Session(bind=eng)
        item = sess.query(docModel).filter(docModel.id==doc_id).first()
        return jsonify({"status":200, "success":True, "data":self.trace_up_parse(item)})

    def trace_up_parse(self,item):
        attrs = dict((i.id, self.item_dict(i)) for i in item.kids)
        rt = []
        for anc in item.ancs:
            result = self.trace_up_parse(anc)
            rt+=result
        attr_id = list(attrs.keys())
        for i in rt:
            for aid in attr_id:
                if aid in i["attrs"]:
                    del attrs[aid]
        thisdt = self.item_dict(item)
        thisdt.update({"attrs": attrs})
        rt = list([thisdt,])+rt
        return rt

    @expose("/read/", methods=["POST"])
    def read_api(self):
        dct = json.loads(request.data)
        doc_id = dct["id"]
        lib = dct["lib"]
        lib_item = self.get_lib(lib)
        eng = ce(lib_item.data)

        sess = Session(bind=eng)
        self.datamodel.session = sess
        item = self.datamodel.get(doc_id)
        kids = list(self.item_dict(i) for i in item.kids)
        parents = list(self.item_dict(i) for i in item.parents)
        dess = list(self.item_dict(i) for i in item.dess)
        ancs = list(self.item_dict(i) for i in item.ancs)
        return jsonify({
            "success": True, "status": 200,
            "data": {"id": item.id,
                     "name": item.name,
                     "names": list(item.names.split(",")),
                     "ctype":item.ctype,
                     "doc": item.doc,#.replace("\n", "<br>"),
                     "level": item.level,
                     "path": item.path,
                     "code": item.code if item.code!="" else "",
                     "kids": kids,
                     "parents": parents,
                     "dess": dess,
                     "ancs": ancs,
                     "lib": lib,
                     "alias":item.alias,
                     }
        })

    @expose("/list_all/", methods=["POST", "GET"])
    def read_all_api(self):
        """
        list all libraries
        :return:
        """
        alllibs = list(
            {"text": i.name, "a": f"/doc/lib/{i.name}/"} for i in self.appbuilder.session.query(libModel).all())
        return jsonify({"status": 200, "success": True if len(alllibs) > 0 else False, "data": alllibs})

    @expose("/lib/<libname>/", methods=["GET"])
    def get_lib_page(self, libname):
        lib_item = self.get_lib(libname)
        eng = ce(lib_item.data)
        sess = Session(bind=eng)
        context = dict({
            "lib":libname,
            "total":sess.query(docModel).count(),
            "classes":sess.query(docModel).filter(docModel.ctype=="class").count(),
            "functions": sess.query(docModel).filter(docModel.ctype == "function").count(),
            "modules": sess.query(docModel).filter(docModel.ctype == "module").count(),
            "attr_relations":sess.query(docGraphModel).count(),
            "inheritance": sess.query(inhGraphModel).count(),
        })
        return self.render_template("lib.html", **context)

    @expose("/code/", methods=["POST"])
    def get_code(self):
        dt = json.loads(request.data)
        lib = dt["lib"]
        name_chain = list(dt["name_chain"].split("."))[1:]
        try:
            obj = __import__(lib)
        except:
            return jsonify({"status":500, "success":False,
                })
        for att in name_chain:
            obj = getattr(obj,att)

        try:
            code  = inspect.getsource(obj)
        except:
            code = "None"
        return jsonify({"status":200, "success":True,
                "data":
                    {"code":code}
                })






from app import app
import argparse
import logging
import os
from doctour.base.utils import startup_log
ap = argparse.ArgumentParser()
ap.add_argument("--port", default=8080, type=int, dest="port")
args = ap.parse_args()

logging.info(startup_log(port=args.port))

if __name__ == "__main__":
    logging.info(f"DocTour Running on port {args.port}")
    app.run(host="0.0.0.0", port=args.port, debug=False)



from flask_appbuilder import Model
from sqlalchemy import Column, Integer, Text
from flask_appbuilder.models.mixins import AuditMixin
from .. import db

class libModel(db.Model):
    __tablename__ = "libs"

    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.Text())
    data = db.Column(db.Text())


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Session": "python_parse.ipynb",
         "basedir": "python_parse.ipynb",
         "most_frequent": "python_parse.ipynb",
         "get_source": "python_parse.ipynb",
         "get_path": "python_parse.ipynb",
         "checks": "python_parse.ipynb",
         "refresh_table": "python_parse.ipynb",
         "docTour": "python_parse.ipynb",
         "parse_lib": "python_parse.ipynb"}

modules = ["base/parse.py"]

doc_url = "https://raynardj.github.io/doctour/"

git_url = "https://github.com/raynardj/doctour/tree/master/"

def custom_doc_links(name): return None



from flask_appbuilder import ModelView,expose
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask import jsonify,request
import json
from doctour.app.models.lib import libModel
from doctour.base.doc import docModel,docGraphModel,inhGraphModel
from doctour.base.parse import docTour, parse_lib
import importlib
import os
import logging
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy import create_engine as ce
import pandas as pd

basedir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))+"/dbs"
Session = sessionmaker()

class libView(ModelView):
    route_base = "/lib"
    datamodel = SQLAInterface(libModel)

    @expose("/add/", methods=["POST"])
    def doc(self):
        dt = json.loads(request.data)
        lib = dt["lib"]
        
        if "dataurl" in dt:
            dataurl = dt["dataurl"]
            self.add_lib_record(lib,dataurl)
        check_lib = list(self.appbuilder.session.query(libModel).filter(libModel.name == lib).all())
        if len(check_lib)>0:
            return jsonify({"status":200,"success":True, "data":{
                "frame":pd.read_sql("SELECT id,name,ctype,alias from docs",con=ce(check_lib[0].data).connect()).to_dict(orient="record")
            }})
        else:
            logging.info(f"parsing {lib}")
            dt, dataurl= parse_lib(lib, import_ = True)
            self.add_lib_record(lib,dataurl)
            rt = {"status": 200, "success": True, "data": {
                "rows": len(dt), "frame": dt.df.to_dict(orient="record")
             }}
        return jsonify(rt)

    def add_lib_record(self,lib,dataurl):
        self.appbuilder.session.add(libModel(name=lib, data=dataurl))
        self.appbuilder.session.commit()

    


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/python_parse.ipynb (unless otherwise specified).

__all__ = ['Session', 'basedir', 'most_frequent', 'get_source', 'get_path', 'checks', 'refresh_table', 'docTour',
           'parse_lib']

# Cell
from .doc import docModel, docGraphModel, inhGraphModel

# Cell
# hide
import pandas as pd
import inspect

import logging
from sqlalchemy import create_engine as ce
from sqlalchemy.orm.session import sessionmaker
import os

# Cell
Session = sessionmaker()
basedir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))+"/app/dbs"

# Cell
def most_frequent(List): return max(set(List), key = List.count)

def get_source(obj):
    try:
        return inspect.getsource(obj)
    except:
        return ""

def get_path(obj):
    try:
        return inspect.getsourcefile(obj)
    except:
        return ""

def checks(obj):
    if inspect.ismodule(obj):return "module"
    if inspect.isclass(obj):return "class"
    if inspect.isfunction(obj):return "function"
    return False

def refresh_table(model,engine):
    """
    Drop the existed table from database, if exist
    In any case create a new table
    """
    table = model.__table__
    if table.exists(engine):
        logging.info(f"dropping existing table: {str(table)}")
        table.drop(engine)
    logging.info(f"creating a new table: {str(table)}")
    table.create(engine)

# Cell
class docTour(object):
    def __init__(self, root_obj, root_name,sess, load_source=False):
        self.docs = dict()
        self.inh_cache = set()
        self.sd_list = list()
        self.root_obj = root_obj
        self.load_source = load_source
        self.sess = sess  # sqla session
        self.doc_parser(self.root_obj, root_name, name_chain=root_name)
        self.vote_name()
        logging.info(f"{len(self.docs)} things found")
        logging.info(f"saving [{root_name}] basics to db")
        self.sess.commit()
        logging.info(f"creating df for [{root_name}]")
        self.df = self.to_df()

    def __len__(self):
        return len(self.docs)

    def mid(self, obj):
        """
        memory address
        """
        return id(obj)

    def to_df(self):
        df = pd.DataFrame(
            list(i.to_dicts("id","name", "doc", "path", "names", "level", "source","ctype","alias") for i in self.docs.values()))
        return df

    def new_doc(self, name_chain, kid, parent, level,ctype):
        """
        New docModel object
        :param name_chain: name chain string
        :param kid: the subject new  obj
        :param parent: parent obj
        :param level: int, level from root obj
        :return: docModel object
        """
        sd = docModel(id=id(kid), name=name_chain.split(".")[-1],
                      names=name_chain, level=level, doc=str(kid.__doc__),
                      ctype = ctype,
                      )
        if ctype=="function":
            path = kid.__code__.co_filename
            if path[-3:] == ".py":
                sd.path = path
                sd.code = inspect.getsource(kid)
        else:
            sd.path = get_path(kid)
        if ctype=="class":
            self.parse_inh(kid)
        if self.load_source:
            sd.source = get_source(kid)

        self.sess.add(sd)
        self.sd_list.append(sd)

        if parent:
            dg = docGraphModel(parent_id=parent.id, kid_id=sd.id)
            self.sess.add(dg)
        return sd

    def parse_inh(self,obj):
        addr = self.mid(obj)
        if addr in self.inh_cache:
            return None
        elif hasattr(obj,"__bases__"):
            bases = obj.__bases__
            if len(bases) == 0: return None

            for b in bases:
                name = obj.__name__
                ig = inhGraphModel(anc_id=id(b), des_id=addr)
                self.sess.add(ig)
                self.parse_inh(b)
                if id(b) not in self.docs:
                    self.doc_parser(b,name,level=1, name_chain=name, parent=None)

            self.inh_cache.add(addr)

    def doc_parser(self, obj, name, level=0, name_chain="", parent=None):
        """
        Parse the sub structure of an object and tracing its documentation
        obj: python class/ object /function
        name: str, name of the object
        level:int, level count from the root obj
        name_chain: str
        parent: docModel,
        """
        addr = self.mid(obj)
        ctype = checks(obj)
        if ctype == False: return None
        if addr in self.docs:
            if type(self.docs[addr]) == docModel:
                sd = self.docs[addr]
                sd.names = sd.names +"," + name_chain
                self.sess.add(sd)
                if parent:
                    dg = docGraphModel(parent_id=parent.id, kid_id=sd.id)
                    self.sess.add(dg)
            return None

        if hasattr(obj, "__doc__"):
            sd = self.new_doc(name_chain, obj, parent, level, ctype)
            self.docs[addr] = sd

        for attr_name in dir(obj):
            sub_obj = getattr(obj, attr_name)
            if checks(sub_obj)==False: continue
            name_chain_ = name_chain + "." + attr_name
            if self.mid(sub_obj) in self.docs:
                sd = self.docs[self.mid(sub_obj)]
                sd.names = sd.names +","+ name_chain_

                if self.load_source:
                    sd.source = get_source(sub_obj)
                self.sess.add(sd)
                if parent:
                    dg = docGraphModel(parent_id=self.docs[addr].id, kid_id=sd.id)
                    self.sess.add(dg)
                continue
            elif "__" not in attr_name:
                if level < 6:
                    try:
                        self.doc_parser(getattr(obj, attr_name), attr_name, level=level + 1, name_chain=name_chain_,
                                        parent=self.docs[addr])
                    except Exception as e:
                        logging.error(f"{name_chain_},\t{e}")

    def vote_name(self):
        logging.info("start voting for most frequent name")
        for sd in self.sd_list:
            name_list = list(n.split(".")[-1] for n in sd.names.split(","))
            sd.name = most_frequent(name_list)
            sd.alias = ",".join(set(name_list))
            self.sess.add(sd)
        logging.info("voting complete")

# Cell
def parse_lib(lib, import_ = True, obj = None):
    path = os.path.join(basedir, f"{lib}.db")
    dataurl = "sqlite:///" + path
    os.system(f"rm {path}")
    print(f"creating SQLite db:\t {dataurl}")
    eng = ce(dataurl)
    sess = Session(bind=eng)

    for m in [docModel,docGraphModel,inhGraphModel]:
        refresh_table(m, engine = eng)

    if import_:
        dt = docTour(__import__(lib), lib, sess)
    else:
        dt = docTour(obj, lib, sess)
    return dt,dataurl




import logging

from flask import Flask
from flask_appbuilder import AppBuilder, SQLA
from doctour import config
"""
 Logging configuration
"""

logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)

app = Flask(__name__)
app.config.from_object(config)
db = SQLA(app)

from flask_appbuilder import IndexView,expose

class indexView(IndexView):
    index_template = "index.html"
    route_base = "/"

    @expose("/",methods=["GET"])
    def index_view(self):
        # inspect environment to offer search suggest
        return 123

appbuilder = AppBuilder(app, db.session,indexview=indexView)


"""
from sqlalchemy.engine import Engine
from sqlalchemy import event

#Only include this for SQLLite constraints
@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    # Will force sqllite contraint foreign keys
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()
"""

from . import views



from flask import render_template
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder import ModelView, ModelRestApi

from .. import appbuilder, db
from .search import searchView, docView
from .lib import libView

"""
    Create your Model based REST API::

    class MyModelApi(ModelRestApi):
        datamodel = SQLAInterface(MyModel)

    appbuilder.add_api(MyModelApi)


    Create your Views::


    class MyModelView(ModelView):
        datamodel = SQLAInterface(MyModel)


    Next, register your Views::


    appbuilder.add_view(
        MyModelView,
        "My View",
        icon="fa-folder-open-o",
        category="My Category",
        category_icon='fa-envelope'
    )
"""

"""
    Application wide 404 error handler
"""


@appbuilder.app.errorhandler(404)
def page_not_found(e):
    return (
        render_template(
            "404.html", base_template=appbuilder.base_template, appbuilder=appbuilder
        ),
        404,
    )

appbuilder.add_view_no_menu(searchView)
appbuilder.add_view(libView, "Libraries")
appbuilder.add_view_no_menu(docView)
db.create_all()



from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="doc_tour",
    version="0.1.0",
    author="raynardj",
    author_email="raynard@rasenn.com",
    description="A web UI for detailed python coding",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages = find_packages(),
    include_package_data=True,
    py_modules=['doctour',],
    scripts = ['doctour/doctour', ],
    package_data={'doctour':['./doctour/templates/*','./doctour/static/*']},
    install_requires = [
        "flask",
        "flask_appbuilder",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import os
from pathlib import Path
from datetime import datetime
import json
import copy

def make_local():
    HOME = Path(os.environ["HOME"])
    LOCAL = HOME/".doctour"
    LOCAL.mkdir(exist_ok=True)
    return LOCAL
    
def logfile(logtype, mode = "w"):
    LOCAL = make_local()
    LOG = LOCAL/"log"
    LOG.mkdir(exist_ok=True)
    LOGFILE = LOG/logtype
    return LOGFILE.open(mode=mode)

def startup_log(**kwargs):
    log = copy.deepcopy(kwargs)
    log.update({
        "time":datetime.now().strftime("%Y%m%d_%H%M%S"),
        "user":os.environ.get("USER"),
        "home":os.environ.get("HOME"),
        })
    logstr = json.dumps(log)
    f = logfile("startapp")
    f.write("\n")
    f.write(logstr)
    f.close()
    return logstr
    


import requests
from doctour.base.utils import logfile
from doctour.base.parse import docTour, parse_lib
import json
from datetime import datetime
import logging
from doctour.base.parse import checks

class exam(object):
    def __init__(self, obj, host = "localhost"):
        self.obj = obj
        self.host = host
        self.port = self.doctour_log["port"]
        self.parse()

    @property
    def doctour_log(self):
        f = logfile("startapp","r")
        jsonstr = f.readlines()[-1]
        return json.loads(jsonstr)

    def get_obj_name(self, obj):
        name = str(obj)
        if hasattr(obj, "__name__"):
            name = obj.__name__
        if hasattr(obj, "__code__"):
            if hasattr(obj, "co_name"):
                name = obj.__code__.co_name
        return name

    def parse(self):
        if checks(self.obj)==False:
            self.obj = self.obj.__class__
        try:
            name = self.get_obj_name(self.obj)
        except:
            name = "object"
        name = f"{name}_{datetime.now().strftime('%H%M%S')}"
        self.name = name
        dt, dataurl = parse_lib(name, import_ = False, obj=self.obj)
        self.dt = dt
        url=f"http://{self.host}:{self.port}/lib/add/"
        dt_ = {"lib":name, "dataurl":dataurl}
        logging.info(url)
        logging.info(str(dt_))
        r = requests.post(url, data=json.dumps(dt_))
        if r.status_code == 200:
            data = r.json()["data"]["frame"]
        self.data  = data
        for d in data:
            if d["name"] == name:
                print(f"please visit http://{self.host}:{self.port}/doc/read/{d['name']}/{d['id']}/")
                return d
        print("Parsing not through")

    def __repr__(self):
        return f"<DocTour:{self.name}>"

